/******************************\
*                              *
* EIGENVALUES AND EIGENVECTORS *
* ---------------------------- *
*                              *
*            PART II           *
*                              *
********************************
*                              *
*       DEFLATION METHOD       *
*        (General case)        *
*                              *
********************************
*                              *
*      Diochnos Dimitris       *
*          (stud1098)          *
*                              *
*      Saltampassis Nick       *
*          (stud1154)          *
*                              *
********************************
*                              *
*       3 - 5 June 2003        *
*                              *
********************************
*                              *
*          REVISIONS           *
*                              *
* 1] Mar 25, 2008              *
*[vars and comments in English]*
*                              *
* 2] Feb 26, 2009              *
* Inserted assertions to verify*
* that starting ellipsis is not*
* symmetric in all directions. *
*                              *
\******************************/

/*
** ! Note: The arrays begin at index number 1 and NOT at index 0.
*/

#include "deflation.h"

/*********************************************************/

int		Deflation (MATRIXPTR array, const double tol, const int max_iterations, 
				   VECTORPTR eigenvalues, MATRIXPTR eigenvectors)
{
	MATRIXPTR		L, L_Inverse, temp_array1, temp_array2;
	VECTORPTR		temp_vector;
	int	const		size = array->size;
	int				i, j, p;
	MATRIXPTR *		Sequence = (MATRIXPTR *) malloc (sizeof (MATRIXPTR) * size);
	double			temp, max;

	temp_vector = Initialize_Vector (size);
	for (i = 1; i <= temp_vector->size; i++)
		Set_Vector (temp_vector, i, (double) 1.0);	// Initial estimation for max eigenvalue.

	Power_Method (array, tol, max_iterations, &temp, temp_vector);	// Calculate Max Eigenvalue.

	Set_Vector (eigenvalues, 1, temp);	// Store MAX eigenvalue.
	for (i = 1; i <= size; i++) {		// Store MAX eigenvector.
		Get_Vector (temp_vector, i, &temp);
		Set_Matrix (eigenvectors, i, 1, temp);
	}

	/* Create empty the matrix sequence that will arise. */
	for (i = 0; i < size; i++)
		Sequence [i] = Initialize_Matrix (size - i, size - i);

	/* Copy Initial matrix. */
	temp_array1 = Initialize_Matrix (size, size);
	for (i = 1; i <= size; i++) {
		for (j = 1; j <= size; j++) {
			Get_Matrix (array, i, j, &temp);
			Set_Matrix (temp_array1, i, j, temp);
		}
	}

	p = 1;
	Get_Matrix (eigenvectors, 1, 1, &max);
	for (i = 2; i <= size; i++) {
		Get_Matrix (eigenvectors, i, 1, &temp);
		if (fabs (temp) > fabs (max)) {
			p = i;
			max = temp;
		}
	}

	/******************************\
	* Simulate unitary matrix with *
	* line interchange.            *
	\******************************/

	if (p != 1) {
		Swap_Lines (temp_array1, 1, p);
		Swap_Columns_Matrix (temp_array1, 1, p);
	}

	L = Initialize_Matrix (size, size);
	L_Inverse = Initialize_Matrix (size, size);

	Swap_Columns_Vector (temp_vector, 1, p);

	for (i = 1; i <= size; i++) {
		Set_Matrix (L, i, i, (double) 1.0);
		Set_Matrix (L_Inverse, i, i, (double) 1.0);
		if (i != 1) {
			Get_Vector (temp_vector, i, &temp);
			Set_Matrix (L, i, 1, 0 - temp);
			Set_Matrix (L_Inverse, i, 1, temp);
		}
	}

	temp_array2 = Initialize_Matrix (size, size);
	
	Multiplication_Matrices (L, temp_array1, temp_array2);
	Multiplication_Matrices (temp_array2, L_Inverse, Sequence [0]);

	/*
	** Free up Memory.
	*/
	Destroy_Matrix (temp_array1);
	Destroy_Matrix (temp_array2);
	Destroy_Matrix (L);
	Destroy_Matrix (L_Inverse);
	Destroy_Vector (temp_vector);

	/*
	** Enter recursive call.
	*/
	Recursive_Deflation (Sequence, tol, max_iterations, eigenvalues, eigenvectors, 0);
	
	/*
	** Complete eigenvectors' co-ordinates.
	*/
	Inflate_Eigenvectors (Sequence, tol, max_iterations, size - 1, size, eigenvalues, eigenvectors);

	/*
	** Complete operation by destroying Sequence[N] Sequence.
	*/
	for (i = size - 1; i >= 0; i--)
		Destroy_Matrix (Sequence [i]);
	free (Sequence);


	return OK;
}

/*********************************************************/
/*****************************************\
* * This function computes the rest (N-1) *
* eigenvalues of the initial matrix.      *
* NOTE that it does NOT fully compute the *
* respective eigenvectors.                *
* * More specifically, at depth depth = k *
* computes the lambda_k eigenvalue and    *
* the last (N - k - 1) co-ordinates of    *
* the respective eigenvector(of lambda_k).*
* (The matrix of eigenvectors is Lower    *
* Triangular when the process is over).   *
* * Finally, the sequence of matrices     *
* that is formed (Sequence [N]) will help *
* us in the next step to compute the      *
* co-ordinates that are missing in the    *
* required positions of the matrix of     *
* eigenvectors.                           *
\*****************************************/

int		Recursive_Deflation (MATRIXPTR * Sequence, const double tol, const int max_iterations, 
		VECTORPTR eigenvalues, MATRIXPTR eigenvectors, const int depth)
{
	MATRIXPTR		L, L_Inverse, temp_array1, temp_array2;
	VECTORPTR		temp_vector;
	int				i, j, p;
	double			temp, max;

	int const size = (int) Sequence [0]->size - depth - 1;

	temp_vector = Initialize_Vector (size);
	for (i = 1; i <= temp_vector->size; i++)
		Set_Vector (temp_vector, i, (double) 1.0);	// Initial estimation for max eigenvalue.

	/* Copy Initial matrix. */
	temp_array1 = Initialize_Matrix (size, size);
	for (i = 2; i <= size + 1; i++) {
		for (j = 2; j <= size + 1; j++) {
			Get_Matrix (Sequence [depth], i, j, &temp);
			Set_Matrix (temp_array1, i - 1, j - 1, temp);
		}
	}

	Power_Method (temp_array1, tol, max_iterations, &temp, temp_vector);	// Calculate Max Eigenvalue.

	Set_Vector (eigenvalues, depth + 2, temp);	// Store MAX eigenvalue.
	for (i = 1; i <= size; i++) {				// Store MAX eigenvector.
		Get_Vector (temp_vector, i, &temp);
		Set_Matrix (eigenvectors, i + depth + 1, depth + 2, temp);
	}

	if (depth == (int) Sequence [0]->size - 2) {
		Set_Matrix (Sequence [Sequence [0]->size - 1], 1, 1, temp_array1->line [0]->column [0]);

		/*
		** Free-Up Memory and complete recursion.
		*/
		Destroy_Matrix (temp_array1);
		Destroy_Vector (temp_vector);

		return	OK;
	}

	p = depth + 2;
	Get_Matrix (eigenvectors, depth + 2, depth + 2, &max);
	for (i = depth + 2; i <= size + 1; i++) {
		Get_Matrix (eigenvectors, i, depth + 2, &temp);
		if (fabs (temp) > fabs (max)) {
			p = i;
			max = temp;
		}
	}

	/******************************\
	* Simulate unitary matrix with *
	* line interchange.            *
	\******************************/

	if (p != depth + 2) {
		Swap_Lines (temp_array1, 1, p - depth - 1);
		Swap_Columns_Matrix (temp_array1, 1, p - depth - 1);
	}

	L = Initialize_Matrix (size, size);
	L_Inverse = Initialize_Matrix (size, size);

	Swap_Columns_Vector (temp_vector, 1, p - depth - 1);

	for (i = 1; i <= size; i++) {
		Set_Matrix (L, i, i, (double) 1.0);
		Set_Matrix (L_Inverse, i, i, (double) 1.0);
		if (i != 1) {
			Get_Vector (temp_vector, i, &temp);
			Set_Matrix (L, i, 1, 0 - temp);
			Set_Matrix (L_Inverse, i, 1, temp);
		}
	}

	temp_array2 = Initialize_Matrix (size, size);
	
	Multiplication_Matrices (L, temp_array1, temp_array2);
	Multiplication_Matrices (temp_array2, L_Inverse, Sequence [depth + 1]);

	/*
	** Free up Memory.
	*/
	Destroy_Matrix (temp_array1);
	Destroy_Matrix (temp_array2);
	Destroy_Matrix (L);
	Destroy_Matrix (L_Inverse);
	Destroy_Vector (temp_vector);


	/*
	** Go Deeper and deeper and ... deeeeeeeper!	:-)
	*/
	//depth++;

	Recursive_Deflation (Sequence, tol, max_iterations, eigenvalues, eigenvectors, depth + 1);

	return	OK;
}

/*********************************************************/
/********************************************\
* The constants row and column declare which *
* which element of eigenvectors we fill in.  *
\********************************************/

int Inflate_Eigenvectors (MATRIXPTR * Sequence, const double tol, const int max_iterations, 
						  const int row, const int column, VECTORPTR eigenvalues, MATRIXPTR eigenvectors)
{
	MATRIXPTR		L_Inverse;
	VECTORPTR		temp_eigenvector, y, b, z;
	double			temp, lambda1, lambda2, max;
	int				i, j, p;

	int const size = (int) Sequence [0]->size;

	temp_eigenvector = Initialize_Vector (size - row + 1);
	z = Initialize_Vector (size - row + 1);
	y = Initialize_Vector (size - row);

	for (i = row; i <= size; i++) {
		Get_Matrix (eigenvectors, i, row, &temp);
		Set_Vector (temp_eigenvector, i - row + 1, temp);
		if (i > row) {
			Get_Matrix (eigenvectors, i, column, &temp);
			Set_Vector (z, i - row + 1, temp);
			Set_Vector (y, i - row, temp);
		}
	}

	b = Initialize_Vector (size - row);

	for (j = 2; j <= Sequence [row - 1]->size; j++) {
		Get_Matrix (Sequence [row - 1], 1, j, &temp);
		Set_Vector (b, j - 1, temp);
	}

	Get_Vector (eigenvalues, row, &lambda1);
	Get_Vector (eigenvalues, column, &lambda2);
	Multiplication_Vectors (b, y, &temp);

	/* Uncomment the following 6 lines to debug. */
	assert (lambda1 != lambda2);
	/*printf ("lambda1 = %.15lf\n", lambda1);
	printf ("lambda2 = %.15lf\n", lambda2);
	if (lambda2 - lambda1 == 0.0) {
		printf ("Eigenvalues appear to be equal! Now exiting.\a\n");
		exit (1);
	}*/
	Set_Vector (z, 1, (double) (temp / (lambda2 - lambda1)));

	/* We have concluded with the computation of z. */

	p = 1;
	Get_Vector (temp_eigenvector, 1, &max);
	for (i = 2; i <= temp_eigenvector->size; i++) {
		Get_Vector (temp_eigenvector, i, &temp);
		if (fabs (temp) > fabs (max)) {
			p = i;
			max = temp;
		}
	}
	Swap_Columns_Vector (temp_eigenvector, 1, p);

	L_Inverse = Initialize_Matrix (Sequence [row - 1]->size, Sequence [row - 1]->size);
	for (i = 1; i <= L_Inverse->size; i++) {
		Set_Matrix (L_Inverse, i, i, (double) 1.0);
		if (i != 1) {
			Get_Vector (temp_eigenvector, i, &temp);
			Set_Matrix (L_Inverse, i, 1, temp / max);
		}
	}
	if (p != 1)
		Swap_Lines (L_Inverse, 1, p);

	Multiplication_Matrix_Vector (L_Inverse, z, temp_eigenvector);

	p = 1;
	Get_Vector (temp_eigenvector, 1, &max);
	for (i = 2; i <= temp_eigenvector->size; i++) {
		Get_Vector (temp_eigenvector, i, &temp);
		if (fabs (temp) > fabs (max)) {
			p = i;
			max = temp;
		}
	}

	for (i = row; i <= size; i++) {
		Get_Vector (temp_eigenvector, i - row + 1, &temp);
		Set_Matrix (eigenvectors, i, column, temp / max);
	}

	Destroy_Matrix (L_Inverse);
	Destroy_Vector (temp_eigenvector);
	Destroy_Vector (z);
	Destroy_Vector (y);
	Destroy_Vector (b);

	if ((column + 1) <= size)
		return	Inflate_Eigenvectors (Sequence, tol, max_iterations, row, column + 1, eigenvalues, eigenvectors);
	else if (row > 1)
		return	Inflate_Eigenvectors (Sequence, tol, max_iterations, row - 1, row, eigenvalues, eigenvectors);
	else
		return OK;
}

/*********************************************************/
