/******************************\
*                              *
*      Diochnos Dimitris       *
*          (stud1098)          *
*                              *
*      Saltampassis Nick       *
*          (stud1154)          *
*                              *
********************************
*                              *
*       3 - 5 June 2003        *
*                              *
********************************
*                              *
*          REVISIONS           *
*                              *
* 1] Mar 25, 2008              *
*[vars and comments in English]*
*                              *
* 2] Feb 26, 2009              *
* Inserted assertions to verify*
* that starting ellipsis is not*
* symmetric in all directions. *
*                              *
\******************************/

#include "matrix.h"

/**************************************************************\
* Note: The arrays begin at index number 1 and NOT at index 0. *
\**************************************************************/

/*
** Functions for manipulation of vectors.
*/

VECTORPTR Initialize_Vector (const int columns)
{
	VECTORPTR	new_vector = (VECTORPTR) NULL;
	int			i;

	if (columns < 1)
		return	new_vector;		// Return NULL pointer to indicate error.

	new_vector = (VECTORPTR) malloc (sizeof (VECTOR));
	new_vector->size = (int) columns;

	new_vector->column = (double *) malloc (sizeof (double) * new_vector->size);
	for (i = 0; i < new_vector->size; i++)
		new_vector->column [i] = (double) 0.0;

	return new_vector;
}

void	Destroy_Vector (VECTORPTR vector)
{
	if (vector == (VECTORPTR) NULL)
		return;

	free (vector->column);
	free (vector);
}

int		Set_Vector (VECTORPTR vector, const int position, const double number)
{
	if (vector == (VECTORPTR) NULL)
		return	SET_NOT_INITIALIZED_VECTOR;

	if ((position > vector->size) || (position < 1))
		return	SET_VECTOR_OUT_OF_BOUNDS;

	vector->column [position - 1] = (double) number;

	return	OK;
}

int		Get_Vector (VECTORPTR vector, const int position, double * result)
{
	if (vector == (VECTORPTR) NULL)
		return	GET_NOT_INITIALIZED_VECTOR;

	if (position > vector->size)
		return	GET_VECTOR_OUT_OF_BOUNDS;

	*result = (double) vector->column [position - 1];

	return	OK;		// Everything was OK.
}

int		Swap_Columns_Vector (VECTORPTR vector, const int column1, const int column2)
{
	double	temp1, temp2;

	if ( ((column1 < 1) || (column1 > vector->size)) || ((column2 < 1) || (column2 > vector->size)) )
		return SWAP_COLUMNS_VECTOR_OUT_OF_BOUNDS;

	Get_Vector (vector, column1, &temp1);
	Get_Vector (vector, column2, &temp2);
	Set_Vector (vector, column1, temp2);
	Set_Vector (vector, column2, temp1);

	return OK;
}

void	Multiplication_Vectors (VECTORPTR v1, VECTORPTR v2, double * result)
{
	int		i;
	double	temp, temp2;

	*result = (double) 0.0;
	for (i = 1; i <= v1->size; i++) {
		Get_Vector (v1, i, &temp);
		Get_Vector (v2, i, &temp2);
		*result += temp * temp2;
	}
}


/*
** Functions for manipulation of arrays.
*/

MATRIXPTR	Initialize_Matrix (const int rows, const int columns)
{
	MATRIXPTR	array = (MATRIXPTR) NULL;
	int			i;

	if ((rows < 1) || (columns < 1))
		return	array;		//Indicate error in bounds.

	array = (MATRIXPTR) malloc (sizeof (MATRIX));
	array->size = (int) rows;
	array->line = (VECTORPTR *) malloc (sizeof (VECTORPTR) * array->size);

	for (i = 0; i < array->size; i++)
		array->line [i] = Initialize_Vector (columns);

	return	array;
}

void	Destroy_Matrix (MATRIXPTR array)
{
	int		i;

	if (array == (MATRIXPTR) NULL)
		return;

	for (i = array->size - 1; i >= 0; i--)
		Destroy_Vector (array->line [i]);

	free (array->line);
	free (array);
}

int		Set_Matrix (MATRIXPTR array, const int row, const int column, const double number)
{
	if (array == (MATRIXPTR) NULL)
		return	SET_NOT_INITIALIZED_MATRIX;

	if ((row < 1) || (row > array->size))
		return	SET_MATRIX_OUT_OF_BOUNDS;

	return	Set_Vector (array->line [row - 1], column, number);
}

int		Get_Matrix (MATRIXPTR array, const int row, const int column, double * result)
{
	if (array == (MATRIXPTR) NULL)
		return	GET_NOT_INITIALIZED_MATRIX;

	if ((row < 1) || (row > array->size))
		return	GET_MATRIX_OUT_OF_BOUNDS;

	return	Get_Vector (array->line [row - 1], column, result);
}

int 	Swap_Lines (MATRIXPTR array, const int line1, const int line2)
{
	VECTORPTR		temp;

	if ( ((line1 < 1) || (line1 > array->size)) || ((line2 < 1) || (line2 > array->size)) )
		return	SWAP_LINES_OUT_OF_BOUNDS;

	temp = array->line [line1 - 1];
	array->line [line1 - 1] = array->line [line2 - 1];
	array->line [line2 - 1] = temp;

	return	OK;
}

int		Swap_Columns_Matrix (MATRIXPTR array, const int column1, const int column2)
{
	int		i;

	if ( ((column1 < 1) || (column1 > array->size)) || ((column2 < 1) || (column2 > array->size)) )
		return SWAP_COLUMNS_MATRIX_OUT_OF_BOUNDS;

	for (i = 0; i < array->size; i++)
		Swap_Columns_Vector (array->line [i], column1, column2);

	return OK;
}

void	Multiplication_Matrices (MATRIXPTR array1, MATRIXPTR array2, MATRIXPTR product)
{
	int		i, j, r;
	double	result, temp, temp2;

	/*
	** Matrix 'product' must be already initialized.
	*/
	for (i = 1; i <= array1->size; i++) {
		for (j = 1; j <= array2->line [0]->size; j++) {
			result = (double) 0.0;
			for (r = 1; r <= array2->size; r++) {
				Get_Matrix (array1, i, r, &temp);
				Get_Matrix (array2, r, j, &temp2);
				result += temp * temp2;
			}
			Set_Matrix (product, i, j, result);
		}
	}
}

void	Multiplication_Matrix_Vector (MATRIXPTR array, VECTORPTR v, VECTORPTR product)
{
	int		i, r;
	double	result, temp, temp2;

	/*
	** Matrix 'product' must be already initialized.
	*/
	for (i = 1; i <= array->size; i++) {
		result = (double) 0.0;
		for (r = 1; r <= array->size; r++) {
			Get_Matrix (array, i, r, &temp);
			Get_Vector (v, r, &temp2);
			result += temp * temp2;
		}
		Set_Vector (product, i, result);
	}
}

int		Get_Column (MATRIXPTR array, const int column, VECTORPTR column_matrix)
{
	double	temp;
	int		i;

	if ((array == (MATRIXPTR) NULL) || (column_matrix == (VECTORPTR) NULL))
		return	GET_COLUMN_MATRIX_UNDEFINED;

	if ((column < 1) || (column > array->size))
		return	GET_COLUMN_MATRIX_OUT_OF_BOUNDS;

	if (array->size != column_matrix->size)
		return	GET_COLUMN_MATRIX_INCOMPATIBLE_TYPES;

	for (i = 1; i <= array->size; i++) {
		Get_Matrix (array, i, column, &temp);
		Set_Vector (column_matrix, i, temp);
	}

	return	OK;
}

int		Set_Column (MATRIXPTR array, const int column, VECTORPTR column_matrix)
{
	double	temp;
	int		i;

	if ((array == (MATRIXPTR) NULL) || (column_matrix == (VECTORPTR) NULL))
		return	SET_COLUMN_MATRIX_UNDEFINED;

	if ((column < 1) || (column > array->size))
		return	SET_COLUMN_MATRIX_OUT_OF_BOUNDS;

	if (array->size != column_matrix->size)
		return	SET_COLUMN_MATRIX_INCOMPATIBLE_TYPES;

	for (i = 1; i <= array->size; i++) {
		Get_Vector (column_matrix, i, &temp);
		Set_Matrix (array, i, column, temp);
	}

	return	OK;
}

void	Print_Matrix (MATRIXPTR array, char * name)
{
	int		rows = array->size, columns = array->line [0]->size;
	int		i, j;

	printf ("%s\n", name);
	for (i = 1; i <= (int) strlen (name); i++)
		printf ("-");
	printf ("\n");
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= columns; j++) {
			printf ("%09.4f  ", array->line [i-1]->column [j-1]);
		}
		printf ("\n");
	}
	printf ("\n");
}
