/******************************\
*                              *
*      Diochnos Dimitris       *
*          (stud1098)          *
*                              *
*      Saltampassis Nick       *
*          (stud1154)          *
*                              *
********************************
*                              *
*       3 - 5 June 2003        *
*                              *
********************************
*                              *
*          REVISIONS           *
*                              *
* 1] Mar 25, 2008              *
*[vars and comments in English]*
*                              *
* 2] Feb 26, 2009              *
* Inserted assertions to verify*
* that starting ellipsis is not*
* symmetric in all directions. *
*                              *
\******************************/


/*
** ! Note: The arrays begin at index number 1 and NOT at index 0.
*/


#include "power.h"

/*********************************************************/

int		Power_Method (MATRIXPTR array, const double tol, const int max_iterations, double * max_eigenvalue, VECTORPTR eigenvector)
{
	int	const		size = (int) array->size;
	int				k = (int) 0, i, j;
	double			lambda0 = (double) 0.0, lambda1, result, temp, temp2, max;
	VECTORPTR		y = Initialize_Vector (array->size), z = Initialize_Vector (array->size);

	/****************************************************\
	* ATTENTION: eigenvector comes as an initialized     *
	* ---------- argument of the function. It represents *
	* the initial estimation of the eigenvector.         *
	* Usually this estimation is the vector: <1,1,...,1> *
	\****************************************************/

	for (i = 1; i <= size; i++)	{
		/* Set initial estimation */
		Get_Vector (eigenvector, i, &result);
		Set_Vector (y, i, (double) result);
	}

	Get_Vector (y, 1, &max);
	for (i = 2; i <= size; i++) {
		Get_Vector (y, i, &result);
		if (fabs (result) > fabs (max))
			max = result;
	}

	for (i = 1; i <= size; i++) {
		Get_Vector (y, i, &result);
		Set_Vector (z, i, result / max);
	}

	while (k <= max_iterations) {
		for (i = 1; i <= size; i++) {
			result = (double) 0.0;
			for (j = 1; j <= size; j++) {
				Get_Matrix (array, i, j, &temp);
				Get_Vector (z, j, &temp2);
				result += (double) temp * temp2;
			}
			Set_Vector (y, i, result);
		}

		Get_Vector (y, 1, &max);
		for (i = 2; i <= size; i++) {
			Get_Vector (y, i, &result);
			if (fabs (result) > fabs (max))
				max = result;
		}

		lambda1 = max;

		if (fabs (lambda1) < ZERO_TOLERANCE) {
			*max_eigenvalue = (double) 0.0;
			for (i = 1; i <= size; i++)
				Set_Vector (eigenvector, i, (double) 0.0);

			Destroy_Vector (y);		// De-Alloccate useless memory.
			Destroy_Vector (z);		// De-Alloccate useless memory.

			return	EIGENVALUE_ZERO;	// Indicate eigenvalue was Zero. Try another initial vector y.
		}

		for (i = 1; i <= size; i++) {
			Get_Vector (y, i, &result);
			Set_Vector (z, i, result / max);
		}

		if (fabs (lambda0 - lambda1) < tol) {
			*max_eigenvalue = (double) lambda1;
			for (i = 1; i <= size; i++)	{
				Get_Vector (z, i, &temp);
				Set_Vector (eigenvector, i, (double) temp);
			}

			Destroy_Vector (y);		// De-Alloccate useless memory.
			Destroy_Vector (z);		// De-Alloccate useless memory.

			return	OK;		// Indicate maximum absolute eigenvalue was found!
		}

		lambda0 = (double) lambda1;
		k++;
	}

	Destroy_Vector (y);
	Destroy_Vector (z);
	max_eigenvalue = (double *) NULL;
	for (i = 1; i <= size; i++)
		Set_Vector (eigenvector, i, (double) 0.0);

	return	POWER_DID_NOT_CONVERGE;
}
